/*
 * Copyright (c) 2015 Tricoire Sebastien 3dsman@free.fr
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 * claim that you wrote the original software. If you use this software
 * in a product, an acknowledgment in the product documentation would be
 * appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 * misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 */

#include "stitchs/OE_fillstitch.h"
#include "curves/OE_pointcurve.h"
#include "curves/OE_joincurve.h"
#include "curves/OE_subcurve.h"
#include "OE_utils.h"
#include <iostream>
#include <stdio.h>

void OE_fillstitch::contour_segment::addIntersection(OE_fillstitch::intersection intersection)
{
	intersections.push_back(intersection);
	if (minXCell == -1)
	{
		if (intersection.xcross)
		{
			minXCell = intersection.gridXPos;
			maxXCell = intersection.gridXPos;
		}
		else
		{
			minYCell = intersection.gridYpos;
			maxYCell = intersection.gridYpos;
		}
	}
	else
	{
		if (intersection.xcross)
		{
			minXCell = mini(minXCell,intersection.gridXPos);
			maxXCell = maxi(maxXCell,intersection.gridXPos);
		}
		else
		{
			minYCell = mini(minYCell,intersection.gridYpos);
			maxYCell = maxi(maxYCell,intersection.gridYpos);
		}
	}
}

OE_fillstitch::patch::patch() : maxYCell(-1)
{
}

void OE_fillstitch::patch::addSegment(contour_segment _segment, int _id)
{
	contour_segments.push_back(_id);
	if (minXCell == -1)
	{
		minXCell = _segment.minXCell;
		minYCell = _segment.minYCell;
		maxXCell = _segment.maxXCell;
		maxYCell = _segment.maxYCell;
	}
	else
	{
		minXCell = mini(minXCell,_segment.minXCell);
		minYCell = mini(minYCell,_segment.minYCell);
		maxXCell = maxi(maxXCell,_segment.maxXCell);
		maxYCell = maxi(maxYCell,_segment.maxYCell);
	}
}

void OE_fillstitch::patch::addCut(int _cut)
{
	cuts.push_back(_cut);
}

OE_fillstitch::cut::cut(int _intersect1, int _intersect2) : contours(4, -1)
{
	intersect1 = _intersect1;
	intersect2 = _intersect2;
}

OE_fillstitch::OE_fillstitch() : curve(nullptr), reverse(false), len(1), width(1), pattern(nullptr)
{
}

OE_fillstitch::OE_fillstitch(OE_thread* thread, float len, float width, OE_pattern* pattern)
{
	curve = new OE_joincurve();
	curve->addDependency(this);
	curve->setClosed(true);
	this->len = len;
	this->width = width;
	this->thread = thread;
	this->pattern = pattern;
	setNeedRefresh();
}

void OE_fillstitch::persist(Pakal::Archive* archive)
{
	OE_stitchs::persist(archive);

	archive->set_type_name("OE_fillstitch");
	archive->value("curve", curve);
	archive->value("reverse", reverse);
	archive->value("len", len);
	archive->value("width", width);
	//archive->value("pattern", pattern);
}


OE_fillstitch::~OE_fillstitch()
{
	std::list<OE_subcurve*>::iterator i = curve->curves.begin();
	 while (i != curve->curves.end())
	 {
		 (*i)->removeDependency(curve);
		 delete *i;
		 i++;
	 }
	if (curve)
	{
		delete curve;
	}
}

/*
void  OE_fillstitch::setCurve(OE_curve* curve)
{
	if (this->curve)this->curve->removeDependency(this);
	this->curve = curve;
	curve->addDependency(this);
	setNeedRefresh();
}
*/

void  OE_fillstitch::setLen(float len)
{
	this->len = len;
	setNeedRefresh();
}
void  OE_fillstitch::setWidth(float width)
{
	this->width = width;
	setNeedRefresh();
}
void  OE_fillstitch::setPattern(OE_pattern* stitch)
{
	pattern=stitch;
	setNeedRefresh();
}

OE_joincurve* OE_fillstitch::getJoincurve()
{
	return curve;
}
float OE_fillstitch::getLen()
{
	return len;
}
float OE_fillstitch::getWidth()
{
	return width;
}
OE_pattern* OE_fillstitch::getPattern()
{
	return pattern;
}

bool OE_fillstitch::check()
{
	return curve && curve->check() && pattern && len;
}

void OE_fillstitch::refresh(float dpi, bool force)
{
	std::vector<vector_2dt> curvePts;
	std::vector<vector_2dt> prevline;
	std::vector<vector_2dt> curline;
	std::vector<std::vector<vector_2d>> gridPts;
	std::vector<intersection> intersections;
	std::vector<int> inflexion;
	std::vector<cut> cuts;
	std::vector<vector_2d> outPts;
	BoundingBox bound;
	vector_2d min;
	vector_2d max;
	OE_pointcurve curve1;
	OE_pointcurve curve2;

	unsigned int x,y = 0;
	if (!check() || !(needRefresh && force))
	{
		return;
	}

	curve->setClosed(true);
	curve->refresh( dpi);
	if (curve->getNpts()<=3)
	{
		return;
	}

	curvePts = curve->discretizeFast(0.1);
	//curvePts = curve->discretizeFast(0.5);
	bound = curve->getBound();
	//curvePts = curve->discretizeRegular(len);

	if (curvePts.size()<2)
	{
		return;
	}

	min = bound.getMin();
	max = bound.getMax();
	min = min + vector_2d(-1, -1);
	max = max + vector_2d(1, 1);

	curve1.addPoint(min);
	curve1.addPoint(min.x+(max.x-min.x)/3, min.y-20);
	curve1.addPoint(min.x+2*(max.x-min.x)/3, min.y+1);
	curve1.addPoint(max.x, min.y);

	curve2.addPoint(min.x, max.y);
	curve2.addPoint(min.x+(max.x-min.x)/3, max.y+5);
	curve2.addPoint(min.x+2*(max.x-min.x)/3, max.y+12);
	curve2.addPoint(max);

	gridPts.resize(len+1);
	for (int i=0; i<len+1; ++i)
	{
		gridPts[i].resize(width+1);
	}

	std::vector<vector_2dt> curve1Pts;
	std::vector<vector_2dt> curve2Pts;
	float curveStep = curve1.getLength(dpi)/len;
	curve1Pts = curve1.discretizeRegular(curveStep);
	curveStep = curve2.getLength(dpi)/len;
	curve2Pts = curve2.discretizeRegular(curveStep);

	vector_2d pt;
	for(unsigned i=0; i<=len; i++)
	{
		for(unsigned j=0; j<=width; j++)
		{
			gridPts[i][j] = curve2Pts.at(i)+(curve1Pts.at(i)-curve2Pts.at(i))/len*j;
/*
			pt = (max-min);
			pt.y = pt.y / width * j;
			pt.x = pt.x / len * i;
			gridPts[i][j] = pt + min;
*/
			outPts.push_back(gridPts[i][j]);
			//outPts.push_back( curve2Pts.at(j)+(curve1Pts.at(j)-curve2Pts.at(j))/width*i);
		}
	}
/*
	for(unsigned i=0; i<width; i++)
	{
		for(unsigned j=0; j<len; j++)
		{
			outPts.push_back( curve2Pts.at(j)+(curve1Pts.at(j)-curve2Pts.at(j))/width*i);
		}
	}
*/

	std::cout << "start  " << std::endl;

	// must find and init to the start cell
	int curXCell=0;
	int curYCell=0;
	bool out = false;
	do
	{
		curYCell = 0;
		do
		{
			out = (PointInTriangle(curvePts[0],
			                       gridPts[curXCell+1][curYCell],
			                       gridPts[curXCell][curYCell],
			                       gridPts[curXCell][curYCell+1]) ||
			       PointInTriangle(curvePts[0],
			                       gridPts[curXCell+1][curYCell],
			                       gridPts[curXCell][curYCell+1],
			                       gridPts[curXCell+1][curYCell+1]));
			curYCell++;
		} while (curYCell<width && !out);
		curXCell++;
	} while (curXCell<len && !out);
	curXCell--;
	curYCell--;

	// find every intersection between the curve and the grid
	char dir = 0;

	bool intersected = false;
	int prevP = curvePts.size()-1;

	vector_2d intersectPt;
	float intersectPosGrid;
	float intersectPosCurve;
	bool up = false;

	for(std::vector<vector_2dt>::size_type p = 0; p < curvePts.size(); p++)
	{
		do
		{
			intersected = false;
			// check if the segment get out the cell and change the cell in consequence
			if (dir!=1 && curXCell>=0 && curYCell>=0 && curYCell<width &&
			    intersect(curvePts[prevP].v, curvePts[p].v,
			              gridPts[curXCell][curYCell], gridPts[curXCell][curYCell+1],
			              intersectPt, intersectPosCurve, intersectPosGrid))
			{
				intersections.push_back(intersection(intersectPt, false, curYCell + intersectPosCurve, p, curXCell));
				dir = 2;
				curXCell--;
				intersected = true;
			}
			if (dir!=2 && curYCell>=0 && curXCell<len && curYCell<width &&
			    intersect(curvePts[prevP].v, curvePts[p].v,
			              gridPts[curXCell+1][curYCell], gridPts[curXCell+1][curYCell+1],
			              intersectPt, intersectPosCurve, intersectPosGrid))
			{
				curXCell++;
				intersections.push_back(intersection(intersectPt, false, curYCell + intersectPosCurve, p, curXCell));
				dir = 1;
				intersected = true;
			}
			if (dir!=3 && curXCell>=0 && curYCell>=0 && curXCell<len &&
			    intersect(curvePts[prevP].v, curvePts[p].v,
			              gridPts[curXCell][curYCell], gridPts[curXCell+1][curYCell],
			              intersectPt, intersectPosCurve, intersectPosGrid))
			{
				intersections.push_back(intersection(intersectPt, true, curYCell, p, curXCell + intersectPosGrid));
				if (up) inflexion.push_back(intersections.size()-1);
				up = false;
				dir = 4;
				curYCell--;
				intersected = true;
			}
			if (dir!=4 && curXCell>=0 && curXCell<len && curYCell<width &&
			    intersect(curvePts[prevP].v, curvePts[p].v,
			              gridPts[curXCell][curYCell+1], gridPts[curXCell+1][curYCell+1],
			              intersectPt, intersectPosCurve, intersectPosGrid))
			{
				curYCell++;
				intersections.push_back(intersection(intersectPt, true, curYCell, p, curXCell + intersectPosGrid));
				if (!up)
				{
					inflexion.push_back(intersections.size()-1);
				}
				up = true;
				dir = 3;
				intersected = true;
			}
		} while (intersected);
		dir = 0;
		prevP = p;
	}
	int nbCross = 0;
	int closestLeft = -1;
	int closestRight = -1;

	if (intersections.size()==0)
	{
		return;
	}

	//for each inflexion point i will find closest intersection on the same grid line
	for(int inflex : inflexion)
	{
		closestLeft = -1;
		closestRight = -1;
		int prevInflex;
		int prev = inflex;
		prevInflex = prev;
		do
		{
			//prevInflex = prev;
			//prev = prev-1%intersections.size();
			prevInflex = prevInflex-1%intersections.size();
		}
		while (intersections[prevInflex].xcross == false ||
		       intersections[prevInflex].gridYpos != intersections[inflex].gridYpos);

		//if (inter == 0) prev = intersections.size()-1;
		for(std::vector<intersection>::size_type p = 0; p < intersections.size(); p++)
		{
			//check every intersection to find the one that are on the same grid line
			if (intersections[p].xcross == true &&
			    intersections[p].gridYpos == intersections[inflex].gridYpos &&
			    p != inflex &&
			    p != prevInflex)
			{
				// check if they are before or after the inflexion point
				if (intersections[p].gridXPos < intersections[inflex].gridXPos)
				{
					if (closestLeft == -1 || intersections[closestLeft].gridXPos < intersections[p].gridXPos)
					{
						closestLeft = p;
					}
					nbCross++;
				}
				else
				{
					if (closestRight == -1 || intersections[closestRight].gridXPos > intersections[p].gridXPos)
					{
						closestRight = p;
					}
				}
			}
		}
		// if we didn't cross the curve twice, we're inside the form
		if (nbCross%2 == 1)
		{
			if (intersections[prevInflex].gridXPos < intersections[inflex].gridXPos)
			{

				cuts.push_back(cut(closestLeft,prevInflex));
				intersections[closestLeft].cutIndex = cuts.size()-1;
				intersections[prevInflex].cutIndex = cuts.size()-1;

				cuts.push_back(cut(inflex,closestRight));
				intersections[inflex].cutIndex = cuts.size()-1;
				intersections[closestRight].cutIndex = cuts.size()-1;
			}
			else
			{
				cuts.push_back(cut(closestLeft,inflex));
				intersections[closestLeft].cutIndex = cuts.size()-1;
				intersections[inflex].cutIndex = cuts.size()-1;

				cuts.push_back(cut(prevInflex,closestRight));

				intersections[prevInflex].cutIndex = cuts.size()-1;
				intersections[closestRight].cutIndex = cuts.size()-1;
			}
		}
		nbCross = 0;
	}

	std::vector<contour_segment> segments (1);
	int startIntersect=0, curIntersect=0, curSegment = 0, curcut=0, curpatch = 0;

	std::vector<patch> patchs;
	std::list<int> patchsStack;
	patchs.push_back(patch());
	patchsStack.push_front(0);

	//loop to start at the first cut
	if (cuts.size())
	{
		while (intersections[startIntersect].cutIndex == -1 && startIntersect<intersections.size())
		{
			startIntersect++;
		}
		segments[0].cutStart = intersections[startIntersect].cutIndex;
	}

	//add the first intersection
	segments[0].addIntersection(intersections[startIntersect]);
	curIntersect = startIntersect+1;


	//loop on every intersections
	do
	{
		if (intersections[curIntersect].cutIndex!=-1)
		{
			curcut = intersections[curIntersect].cutIndex;
			segments[curSegment].cutEnd = curcut;
			//add segment to current patch and segments array then create a new one

			//std::cout<<"    patch "<<curpatch<<" get "<< curSegment <<std::endl;
			patchs[curpatch].addSegment(segments[curSegment], segments.size()-1);
			segments.push_back(contour_segment());
			curSegment = segments.size() -1;

			segments[curSegment].addIntersection(intersections[curIntersect]);

			// if we already marked this cut we can pop the patch out of the stack and mark contours indexes
			if (cuts[curcut].contours[0]!=-1)
			{
				cuts[curcut].contours[1] = curSegment-1;
				cuts[curcut].contours[3] = curSegment;
				patchsStack.pop_front();
				// if the stack is empty we are going on a new pach
				if (patchsStack.empty())
				{
					patchs.push_back(patch());
					patchsStack.push_front(patchs.size()-1);
				}

			}
			else
			{
				//else mark the cut
				cuts[curcut].contours[0] = curSegment;
				cuts[curcut].contours[2] = curSegment-1;
				//then init a new patch
				patchs.push_back(patch());
				patchsStack.push_front(patchs.size()-1);

			}
			// mark the segment cutStart and cutEnd
			segments[curSegment].cutStart = curcut;
			segments[curSegment-1].cutEnd = curcut;
			curpatch = patchsStack.front();

		}
		//increment the current intersection
		curIntersect = (curIntersect+1)%intersections.size();
		//add the curent intersection to the current segment
		segments[curSegment].addIntersection(intersections[curIntersect]);

	} while (curIntersect!=startIntersect);
	//push back the last segment and mark his start
	//segments.push_back(segment);
	patchs[0].addSegment(segments[segments.size()-1], segments.size()-1);
	if (cuts.size())
	{
		segments[segments.size()-1].cutEnd = intersections[startIntersect].cutIndex;
	}


	// -----------  show result  --------------

	//outPts.clear();
	int curCut = 0;
	int cursegment = -1;
	outPts.clear();
	for(std::vector<patch>::size_type curPatch = 0; curPatch < patchs.size(); curPatch++)
	{
		for(std::vector<contour_segment>::size_type segment = 0; segment < patchs[curPatch].contour_segments.size(); segment++)
		{
			for(std::vector<intersection>::size_type p = 0; p < segments[patchs[curPatch].contour_segments[segment]].intersections.size(); p++)
			{
				outPts.push_back(segments[patchs[curPatch].contour_segments[segment]].intersections[p].pt);
			}
		}
		//outPts.push_back(segments[patchs[curPatch].contour_segments[0]].intersections[0].pt);
	}

	pts.clear();

	if(maxLen>0)
		pts = stitchMaxLen(outPts, maxLen);
	else
		pts = outPts;
	//std::cout<<"end  "<<std::endl;

	needRefresh = false;
}

void OE_fillstitch::delDependency(OE_base * object)
{
	if (curve == object)
	{
		curve = 0;
	}
}

void OE_fillstitch::refreshDependency()
{
	if (curve)
	{
		curve->addDependency(this);
		curve->refreshDependency();
	}
}

